# coding=utf-8
from __future__ import unicode_literals

import inspect

from django.core.exceptions import ValidationError
from django.utils.functional import cached_property

from .errors import ExtensibleValidationError


class CleanMixin(object):
    @cached_property
    def validations(self):
        return ValidationManager(self)

    def clean(self):
        self.validations()


class ValidationManager(object):
    def __init__(self, model, clean_methods=None):
        self.errors = []
        self.model = model
        self.clean_methods = clean_methods or [method for method in dir(model) if method.startswith('clean_') and callable(getattr(model, method))]

    def __call__(self):
        self.full_clean()

    def full_clean(self):
        self.errors = []
        self.collect_validation_errors()
        if self.errors:
            raise ExtensibleValidationError(self.errors)

    def collect_validation_errors(self):
        for name in self.clean_methods:
            method = getattr(self.model, name)
            try:
                if inspect.isgeneratorfunction(method):
                    self.errors.extend(method())
                else:
                    error = method()
                    if isinstance(error, ValidationError):
                        self.append_error(error, name)
            except ValidationError as ex:
                self.append_error(ex, name)

    def append_error(self, error, clean_name):
        if not hasattr(error, 'error_dict'):
            field_name = self.get_field_name(clean_name)
            error = ValidationError({field_name: error})
        self.errors.append(error)

    # noinspection PyMethodMayBeStatic
    def get_field_name(self, clean_name):
        field_name = clean_name.replace('clean_', '')
        return field_name



# -*- coding: utf-8 -*-

"""Unit test package for django-model-cleanup."""

# Having tests in a separate folder is a safe choice
# You can have them mixed in the src but at some point you will get
# problems with test discovery activating lazy objects and other import time
# headaches. And if you fail to filter test during packaging you will get
# similar problems in production.



# coding=utf-8

from django.core.exceptions import ValidationError

from tests.app import models


def test_model_has_validations():
    item = models.SomeModel()
    assert hasattr(item, "validations")


def test_model_errors():
    item = models.SomeModel()
    try:
        item.full_clean()
    except ValidationError as ex:
        assert ex.message_dict == {
            'bar': ['Bar is wrong cause 2 > 1!', 'Bar legacy error 7 > 5!'],
            'foo': ['Foo is bad'],
        }



# coding=utf-8
"""Top-level package for Django Model Validation Cleanup."""
from .errors import ExtensibleValidationError  # noqa F401
from .models import CleanMixin  # noqa F401

__author__ = """Janusz Skonieczny"""
__email__ = 'js+pypi@bravelabs.pl'
__version__ = '0.1.1'



# coding=utf-8
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from django_model_cleanup import CleanMixin, ExtensibleValidationError


class SomeModel(CleanMixin, models.Model):
    lorem = models.CharField(max_length=10, blank=True)

    def clean_foo(self):
        raise ValidationError('Foo is bad')

    def clean_bar(self):
        raise ExtensibleValidationError({'bar': _('Bar is wrong cause %s > %s!')}, code='bar', params=(2, 1))

    def clean_legacy(self):
        # We can't init ValidationError as one-liner, cause dict + params are not compatible
        # We need to wrap a message in ValidationError and put that in dict indicating a field
        msg = _('Bar legacy error %s > %s!')
        error = ValidationError(msg, code='bar', params=(7, 5))
        raise ValidationError({'bar': error})

    # This is not longer required:
    def legacy_clean(self):
        errors = []
        try:
            self.clean_foo()
        except ValidationError as ex:
            errors.append(ex)
        errors = []
        try:
            self.clean_bar()
        except ValidationError as ex:
            errors.append(ex)
        errors = []
        try:
            self.clean_legacy()
        except ValidationError as ex:
            errors.append(ex)
        if errors:
            raise ValidationError(errors)



# coding=utf-8
from __future__ import unicode_literals

from collections import OrderedDict

import six
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.utils.encoding import force_text


class ExtensibleValidationError(ValidationError):
    def __init__(self, message, code=None, params=None):
        """
        The `message` argument can be a single error, a list of errors, or a
        dictionary that maps field names to lists of errors. What we define as
        an "error" can be either a simple string or an instance of
        ValidationError with its message attribute set, and what we define as
        list or dictionary can be an actual `list` or `dict` or an instance
        of ValidationError with its `error_list` or `error_dict` attribute set.
        """

        # PY2 can't pickle naive exception: http://bugs.python.org/issue1692335.
        super(ValidationError, self).__init__(message, code, params)

        if isinstance(message, ValidationError):
            if hasattr(message, 'error_dict'):
                message = message.error_dict
            # PY2 has a `message` property which is always there so we can't
            # duck-type on it. It was introduced in Python 2.5 and already
            # deprecated in Python 2.6.
            elif not hasattr(message, 'message' if six.PY3 else 'code'):
                message = message.error_list
            else:
                message, code, params = message.message, message.code, message.params

        if isinstance(message, dict):
            self.error_dict = {}
            for field, messages in message.items():
                if not isinstance(messages, ExtensibleValidationError):
                    # Let's pass on code and params along with messages from dict
                    messages = ExtensibleValidationError(messages, code, params)
                self.error_dict[field] = messages.error_list

        elif isinstance(message, list):
            self.error_list = []
            for item in message:
                if not isinstance(item, ExtensibleValidationError):
                    # Normalize plain strings to instances of ValidationError.
                    item = ExtensibleValidationError(item, code, params)
                if hasattr(item, 'error_dict') and hasattr(self, 'error_list'):
                    # Convert self from list to dict and prepare for dict update later on
                    self.error_dict = OrderedDict()
                    if self.error_list:
                        self.error_dict[NON_FIELD_ERRORS] = self.error_list
                    del self.error_list
                if hasattr(self, 'error_list'):
                    # Extend error_list with passed in item.error_list
                    self.error_list.extend(item.error_list)
                elif hasattr(item, 'error_list'):
                    # Extend __all__ errors with passed in item.error_list
                    self.error_dict.setdefault(NON_FIELD_ERRORS, []).extend(item.error_list)
                else:
                    # Concat error dictionaries
                    for field, errors in item.error_dict.items():
                        self.error_dict.setdefault(field, []).extend(errors)
        else:
            self.message = message
            self.code = code
            self.params = params
            self.error_list = [self]

    @property
    def messages(self):
        if hasattr(self, 'error_dict'):
            return sum(OrderedDict(self).values(), [])
        return list(self)

    def __iter__(self):
        if hasattr(self, 'error_dict'):
            for field, errors in self.error_dict.items():
                yield field, list(ValidationError(errors))
        else:
            for error in self.error_list:
                message = error.message
                if error.params:
                    message %= error.params
                yield force_text(message)



# -*- coding: utf-8
from __future__ import absolute_import, unicode_literals

from django.conf.urls import include, url

urlpatterns = [
    url(r'^', include('django_model_cleanup.urls', namespace='django_model_cleanup')),
]



# coding=utf-8
"""
Here We're testing assumptions about django's ValidationError class

ValidationError init is a hackish mess at best. It's hard to grasp what it does exactly and why.
But we'll no there to judge but to make sure we have a place to check what it does and what to expect.
"""
from __future__ import unicode_literals

import pytest
from django.core.exceptions import ValidationError
from django.db.models.fields.related import ForeignKey
from django.forms.utils import ErrorDict, ErrorList
from django.utils import translation
from django.utils.translation import override

# This will be formatted differently on py27
EN_MESSAGE = "Foo instance with Bar %r does not exist." % 'Baz'
PL_MESSAGE = "Foo z polem Bar o wartości %r nie istnieje." % 'Baz'
FR_MESSAGE = "L'instance Foo avec %r dans Bar n'existe pas." % 'Baz'


def test_init_with_string():
    # Start simple
    v = ValidationError('A bad value here!')
    assert v.messages == ['A bad value here!']
    assert v.error_list == [v]
    assert hasattr(v, 'error_dict') is False


def test_init_with_string_and_params():
    # Add formatting
    message, params = 'Not correct %d!=%d %s!', (1, 2, 'bam!')
    v = ValidationError(message, params=params)
    assert v.messages == ['Not correct 1!=2 bam!!']
    assert v.error_list == [v]
    assert v.message == message
    assert v.params == params


def test_init_with_dict():
    # dict will trigger recursive init of multiple ValidationErrors
    v = ValidationError({'field': 'A test error message'})
    # will have special attribute
    assert hasattr(v, 'error_dict')
    error = v.error_dict

    # it will mimic original dict
    assert 'field' in error
    field_errors = error['field']

    # but instead of single error we'll have a list
    assert isinstance(field_errors, list)
    assert isinstance(field_errors[0], ValidationError)

    # where we'll have a list errors again
    assert field_errors[0].messages == ['A test error message']
    # this error_list is a one element list of [self]
    assert field_errors[0].error_list == [field_errors[0]]

    # all this will be flattened with messages property
    assert v.messages == ['A test error message']
    # and normalize messages in a dict as s list
    assert v.message_dict == {'field': ['A test error message']}


def test_init_with_string_list():
    messages = ['Error A', 'Error B']
    v = ValidationError(messages)
    # no error_dict
    assert hasattr(v, 'error_dict') is False

    # but messages are here
    assert v.messages == messages

    # and we have error_list
    assert hasattr(v, 'error_list')
    assert isinstance(v.error_list, list)

    # which has 2 elements, as an initial string list
    assert isinstance(v.error_list[0], ValidationError)
    assert isinstance(v.error_list[1], ValidationError)

    # which in turn have both a one element of list of error message
    assert v.error_list[0].messages == ['Error A']
    assert v.error_list[1].messages == ['Error B']

    # And they contain itself in error_list
    assert v.error_list[0].error_list == [v.error_list[0]]
    assert v.error_list[1].error_list == [v.error_list[1]]


def test_init_with_dict_and_params():
    v = ValidationError({'field': 'Not true: %s>%s'}, params=(1, 2))
    # Sorry params do not work with dict
    assert v.messages == ['Not true: %s>%s']
    # TODO: Make sure params work with dict initialization

    # to overcome this we need to pass a validation error instance
    inner = ValidationError('Not true: %s>%s', params=(1, 2))
    v = ValidationError({'field': inner})
    # Sorry params do not work with dict
    assert v.messages == ['Not true: 1>2']
    # this is hack'y
    # but we still have an error dict and message dict, few!
    assert v.error_dict == {'field': [inner]}
    assert v.message_dict == {'field': ['Not true: 1>2']}


def test_init_with_list_and_params():
    v = ValidationError(['Not true: %s>%s', 'Yep, still wrong %s>%s'], params=(1, 2))
    # Sorry params do not work with list also
    assert v.messages == ['Not true: %s>%s', 'Yep, still wrong %s>%s']
    # TODO: Make sure params work with list initialization


@pytest.fixture
def polish_language(settings):
    settings.LANGUAGE_CODE = 'pl'


def get_translated_validation_error(**kwargs):
    """We'll use some internal django message that is supposed to be translated"""
    message = ForeignKey.default_error_messages['invalid']
    params = {'model': 'Foo', 'field': 'Bar', 'value': 'Baz'}
    params.update(kwargs)
    return ValidationError(message, params=params)


def test_init_with_translated_text(polish_language):
    # Add formatting and language
    v = get_translated_validation_error()
    # when asked for message will be translated
    assert v.messages == [PL_MESSAGE]
    # but is stored as lazy object
    assert v.message.__class__.__name__ == '__proxy__'
    # so changing the language is no problem
    with translation.override('fr'):
        assert v.messages == [FR_MESSAGE]


def test_init_with_dict_and_params_and_translation(polish_language):
    v = get_translated_validation_error()

    # Params encapsulation still works with translated strings. Cool.
    v = ValidationError({'some_field': v})
    # when asked for message will be translated
    assert v.messages == [PL_MESSAGE]
    assert v.message_dict == {'some_field': [PL_MESSAGE]}

    # is still stored as lazy object
    assert v.error_dict['some_field'][0].message.__class__.__name__ == '__proxy__'

    # and changing the language has still a desired effect
    with translation.override('fr'):
        assert v.messages == [FR_MESSAGE]
        assert v.message_dict == {'some_field': [FR_MESSAGE]}


def test_concat_errors_simple():
    # you can't raise multiple validation exceptions
    # and giving them to user one after another is a bad UX
    # so we would like to concat them, passing a list to init works
    v1 = ValidationError('foo')
    v2 = ValidationError('bar')
    error = ValidationError([v1, v2])
    assert error.messages == ['foo', 'bar']


def test_concat_errors_with_field_keys():
    # but want if want to concat errors indicating a field?
    v1 = ValidationError({'a': 'foo'})
    v2 = ValidationError({'b': 'bar'})
    error = ValidationError([v1, v2])
    assert error.messages == ['foo', 'bar']
    # no error dict, huh?
    assert not hasattr(error, 'error_dict ')
    # but there is error_list with them, right?
    # no
    assert error.error_list[0] != v1
    # maybe its somewhat similar copy, it should have error_dict
    # no
    assert not hasattr(error.error_list[0], 'error_dict')
    # TODO: Preserve error_dict {field: messages} with concatenations


def test_concat_errors_with_field_and_params():
    # so what about params?
    v1 = ValidationError({'field': 'Not true: %s>%s'}, params=(1, 2))
    v2 = ValidationError({'field': 'Also Not true: %s>%s'}, params=(1, 3))
    error = ValidationError([v1, v2])
    # sorry, we already know that params do no work with dict init, see test_init_with_dict_and_params
    assert error.messages == ['Not true: %s>%s', 'Also Not true: %s>%s']

    # but there was a hack
    v1 = ValidationError('Not true: %s>%s', params=(1, 2))
    v2 = ValidationError('Also Not true: %s>%s', params=(1, 3))
    v1 = ValidationError({'field': v1})
    v2 = ValidationError({'field': v2})
    error = ValidationError([v1, v2])
    # yay, it still works
    assert error.messages == ['Not true: 1>2', 'Also Not true: 1>3']
    # but, still no error_dict
    assert not hasattr(error, 'error_dict ')
    assert not hasattr(error.error_list[0], 'error_dict')


def test_concat_translated_errors(polish_language):
    # lets try concat error with translated text
    v = get_translated_validation_error()
    error = ValidationError([v])

    # yay, works
    assert error.messages == [PL_MESSAGE]
    translation.activate('fr')
    assert error.messages == [FR_MESSAGE]


def test_concat_translated_errors_and_fields(polish_language):
    # lets try concat error with translated text
    v = get_translated_validation_error()
    v = ValidationError({'boo': v})
    error = ValidationError([v])

    # yup, still works
    assert error.messages == [PL_MESSAGE]
    translation.activate('fr')
    assert error.messages == [FR_MESSAGE]
    # but, still no error_dict
    assert not hasattr(error, 'error_dict ')
    assert not hasattr(error.error_list[0], 'error_dict')


def test_forms_error_dict():
    # the ultimate end game is to show these to the user
    # django wraps validation errors with ErrorDict consisting of ErrorList's
    # this is done in the BaseModelForm._post_clean method
    # which in turn calls _update_error that overrides message text if needed
    # and calls add_error with a single ValidationError instance
    # which must have an error_dict
    # TODO: Final ValidationError must have an error_dict with __all__ where field as originally missing
    # for each field and __all__ a ErrorList instance will be created
    # and extended with ValidationError instance error_list for given key.

    v = get_translated_validation_error()
    v = ValidationError({'boo': v, 'zap': [ValidationError('An error here!'), ValidationError('Yet another one!')]})

    form_errors = ErrorDict()
    # lest simulate add_error here
    for field, errors in v.error_dict.items():
        form_field_errors = form_errors.setdefault(field, ErrorList())
        form_field_errors.extend(errors)

    assert form_errors == {'boo': [EN_MESSAGE], 'zap': ['An error here!', 'Yet another one!']}
    # Let's find out how language change will affect the result
    with override('fr'):
        assert form_errors == {'boo': [FR_MESSAGE], 'zap': ['An error here!', 'Yet another one!']}
